; v1.1  8-Sep-03, updated by ES
;
; - fixed several PCX issues (header, page filling, print problems)
; - set page size closer to A3/A4
; - converted to 8086 (for old machines!)
;
;***************************************************************************
;
;  PCX.DRV v1.01, written by Matthias Hartl (maze@altmuehlnet.de)
;  last updated : 2002/07/08, 00:00 CEST
;
;  PCX.DRV is a virtual Printer driver for Protel Schplot/Traxplot/Easyplot,
;  which generates Z-Soft PCX 3.0 compliant black/white bitmaps with
;  resolutions ranging from 100 to 1200 DPI.
;
;  For proper functioning, output *MUST* be redirected to a file!
;
;***************************************************************************

IFDEF DPI100
MULT			EQU	1
PRINTER_STRING		EQU	' PCX 100DPI A3/4'
ENDIF

IFDEF DPI300
MULT			EQU	3
PRINTER_STRING		EQU	' PCX 300DPI A3/4'
ENDIF

IFDEF DPI600
MULT			EQU	6
PRINTER_STRING		EQU	' PCX 600DPI A3/4'
ENDIF

IFDEF DPI1200
MULT			EQU	12
PRINTER_STRING		EQU	'PCX 1200DPI A3/4'
ENDIF

HDPI			EQU	(MULT * 100)
VDPI			EQU	(MULT * 100)
HRES_NORMAL		EQU	(MULT * 1169)
VRES_NORMAL		EQU	(MULT * 827)
HRES_REDUCED		EQU	(MULT * 1654)
VRES_REDUCED		EQU	(MULT * 1169)

PCX_MANUFACTURER_ZSOFT	EQU	10
PCX_VERSION_3_0		EQU	5
PCX_ENCODING_RLE	EQU	1
PCX_HEADER_SIZE		EQU	128
PCX_COLOR_PALETTE	EQU	1

	.8086

CODE	SEGMENT BYTE PRIVATE
	ASSUME CS:CODE
	ORG 0000H

BEGIN:

;***************************************************************************
;
;  The following instructions form a well-known jump table. Order as well
;  as encoding is critical and may not be changed. Every jump must encode
;  to a 3 byte instruction (e.g. E9 AD 01), otherwise the driver will not
;  be accepted!
;
;***************************************************************************

	JMP	GET_DRIVER_TYPE
	JMP	GET_UNKNOWN
	JMP	GET_HDPI
	JMP	GET_VDPI
	JMP	GET_HRES
	JMP	GET_VRES
	JMP	GET_NUM_LINES
	JMP	GET_PAPER_FEED_TYPE
	JMP	PRINT
	JMP	SAVE_PARAMETERS
	JMP	PUT_PIXEL

PRINTER_TYPE	DB	16, PRINTER_STRING
DRIVER_TYPE	DB	24, 'Protel Printer Driver   '

	JMP	SELECT_SCALE

HRES			DW	HRES_NORMAL
VRES			DW	VRES_NORMAL
NUM_LINES		DW	?
NUM_PIXELS		DW	?
STACK_POINTER		DW	?
WRITE_DATA_ADDRESS	DW	?
PIXEL_DATA_OFFSET	DW	?
PIXEL_DATA_SEGMENT	DW	?
ENVVAR			DB	'PCXDRV='
SWITCH_POS		DB	'POS'
SWITCH_NEG		DB	'NEG'
POS			DB	1
PARSED			DB	0
LINE_CNT		DW	0
GET_DATA_ADDRESS	DW	?

HEADER:
HDR_MANUFACTURER	DB	PCX_MANUFACTURER_ZSOFT
HDR_VERSION		DB	PCX_VERSION_3_0
HDR_ENCODING		DB	PCX_ENCODING_RLE
HDR_BITS_PER_PIXEL	DB	01H
HDR_XMIN		DW	0000H
HDR_YMIN		DW	0000H
HDR_XMAX		DW	?
HDR_YMAX		DW	?
HDR_HDPI		DW	?
HDR_VDPI		DW	?
HDR_COLOR_MAP		DB	48 DUP (00H)
HDR_RESERVED		DB	00H
HDR_NUM_PLANES		DB	01H
HDR_BYTES_PER_LINE	DW	?
HDR_PALETTE_INFO	DW	PCX_COLOR_PALETTE
HDR_FILLER		DB	58 DUP (00H)

;***************************************************************************
;
	SELECT_SCALE PROC NEAR
;
;	set internal parameters depending on the selected scale
;
;***************************************************************************

	PUSH	BP
	MOV	BP, SP

	CALL	GET_DATA_DISPLACEMENT

	MOV	AX, SS:[BP+4]
	CMP	AX, 2
	JE	SET_REDUCED

	MOV	CS:[SI+HRES], HRES_NORMAL
	MOV	CS:[SI+VRES], VRES_NORMAL
	JMP	SHORT DONE

SET_REDUCED:
	MOV	CS:[SI+HRES], HRES_REDUCED
	MOV	CS:[SI+VRES], VRES_REDUCED

DONE:	MOV	SP, BP
	POP	BP
	RET	2

SELECT_SCALE ENDP

;***************************************************************************
;
	GET_DRIVER_TYPE PROC NEAR
;
;	report driver type (1 = printer driver)
;
;***************************************************************************

	MOV	AX, 1

	RET

GET_DRIVER_TYPE ENDP

;***************************************************************************
;
	GET_UNKNOWN PROC NEAR
;
;	purpose of this procedure is still unknown; must return '100' or
;	printer driver is rejected
;
;***************************************************************************

	MOV	AX, 100

	RET

GET_UNKNOWN ENDP

;***************************************************************************
;
	GET_HDPI PROC NEAR
;
;	report printer's horizontal resolution in dots per inch
;
;***************************************************************************

	MOV	AX, HDPI

	RET

GET_HDPI ENDP

;***************************************************************************
;
	GET_VDPI PROC NEAR
;
;	report printer's vertical resolution in dots per inch
;
;***************************************************************************

	MOV	AX, VDPI

	RET

GET_VDPI ENDP

;***************************************************************************
;
	GET_HRES PROC NEAR
;
;	report printer's total horizontal resolution (hor. paper size)
;
;***************************************************************************

	PUSH	SI

	CALL	GET_DATA_DISPLACEMENT
	MOV	AX, CS:[SI+HRES]

	POP	SI
	RET

GET_HRES ENDP

;***************************************************************************
;
	GET_VRES PROC NEAR
;
;	report printer's total vertical resolution (ver. paper size)
;
;***************************************************************************

	PUSH	SI

	CALL	GET_DATA_DISPLACEMENT
	MOV	AX, CS:[SI+VRES]

	POP	SI
	RET

GET_VRES ENDP

;***************************************************************************
;
	GET_BPL PROC NEAR
;
;	calculate number of bytes for each raster line
;
;***************************************************************************

	PUSH	BX

	CALL	GET_HRES
	MOV	BX, AX
	SHR	AX, 1
	SHR	AX, 1
	SHR	AX, 1
	AND	BX, 7
	JZ	DONE8
	INC	AX

DONE8:	POP	BX
	RET

GET_BPL ENDP

;***************************************************************************
;
	GET_NUM_LINES PROC NEAR
;
;	calculate number of raster lines fitting into the printer
;	buffer at once
;
;***************************************************************************

	PUSH	BP
	MOV	BP, SP

	CALL	GET_BPL
	MOV	BX, AX
	MOV	AX, SS:[BP+4]
	XOR	DX, DX
	DIV	BX

	MOV	SP, BP
	POP	BP
	RET	2

GET_NUM_LINES ENDP

;***************************************************************************
;
	GET_PAPER_FEED_TYPE PROC NEAR
;
;	report paper feed type (0 = paged, 1 = continous)
;
;***************************************************************************

	MOV	AX, 0

	RET

GET_PAPER_FEED_TYPE ENDP

;***************************************************************************
;
	PRINT PROC NEAR
;
;	encode printer buffer and put it onto output stream
;
;***************************************************************************

	PUSH	BP
	PUSH	ES
	MOV	BP, SP

	CALL	GET_DATA_DISPLACEMENT
	MOV	CS:[SI+STACK_POINTER], BP

	CALL	GET_DRIVER_SETTINGS

	MOV	AX, SS:[BP+6]
	CMP	AX, 1
	JE	DO_HEADER
	CMP	AX, 3
	JE	DO_HEADER
	JMP	SHORT SKIP_HEADER

DO_HEADER:
	CALL	EJECT
	CALL	WRITE_HEADER

	MOV	AX, CS:[SI+VRES]
	MOV	CS:[SI+LINE_CNT], AX

SKIP_HEADER:
	MOV	DX, CS:[SI+NUM_LINES]
	CALL	GET_PIXEL_DATA_ADDR

	MOV	AX, OFFSET GET_BYTE
	MOV	CS:[SI+GET_DATA_ADDRESS], AX

	CALL	NEXT_LINE

	MOV	AX,[BP+6]
	CMP	AX,2
	JE	DONE6
	CMP	AX,3
	JE	DONE6
	JMP	SHORT DONE7

DONE6:	CALL	EJECT

DONE7:	XOR	AX, AX
	MOV	SP, BP
	POP	ES
	POP	BP
	RET	4

;	OUTPUT BUFFER

NEXT_LINE:
	XOR	BX, BX
	MOV	CX, 0
	PUSH	DI

NEXT_DATA:
	MOV	AX, CS:[SI+GET_DATA_ADDRESS]
	ADD	AX, SI
	CALL	AX

	CMP	CS:[SI+POS], 1
	JNE	NEGATIVE
	NOT	AL

NEGATIVE:
	CMP	AL, BL
	JNE	END_OF_RUN

	INC	BH
	CMP	BH, 03FH
	JNE	CHECK_EOL

	PUSH	BX
	CALL	WRITE_RLE

	XOR	BX, BX
	JMP	SHORT CHECK_EOL

END_OF_RUN:
	PUSH	BX
	CALL	WRITE_RLE

	MOV	BL, AL
	MOV	BH, 1

CHECK_EOL:
	INC	DI
	ADD	CX, 8
	CMP	CX, CS:[SI+NUM_PIXELS]
	JB	NEXT_DATA

	PUSH	BX
	CALL	WRITE_RLE

	CALL	GET_BPL
	AND	AX, 1		; IF ODD
	JZ	DONE_LINE

	SUB	BX,BX		; COMPLETE LINE WITH A ZERO
	PUSH	BX
	CALL	WRITE_DATA

DONE_LINE:
	POP	DI
	CALL	GET_BPL
	ADD	DI, AX

	DEC	WORD PTR CS:[SI+LINE_CNT]

	DEC	DX
	JNZ	NEXT_LINE

	RET

;	EJECT PAGE

EJECT:	MOV	DX, CS:[SI+LINE_CNT]
	OR	DX, DX
	JE	EJECT1
	MOV	AX, OFFSET GET_ZERO
	MOV	CS:[SI+GET_DATA_ADDRESS], AX
	CALL	NEXT_LINE
EJECT1:	RET

;	GET NEXT CHARACTER

GET_BYTE:
	MOV	AL, ES:[DI]
	RET

GET_ZERO:
	MOV	AL, 0
	RET

PRINT ENDP

;***************************************************************************
;
	SAVE_PARAMETERS PROC NEAR
;
;***************************************************************************

	PUSH	BP
	MOV	BP, SP

	CALL	GET_DATA_DISPLACEMENT

	MOV	AX, SS:[BP+4]
	MOV	CS:[SI+NUM_LINES], AX
	MOV	AX, SS:[BP+6]
	MOV	CS:[SI+NUM_PIXELS], AX
	MOV	AX, SS:[BP+8]
	MOV	CS:[SI+WRITE_DATA_ADDRESS], AX
	MOV	AX, SS:[BP+12]
	MOV	CS:[SI+PIXEL_DATA_OFFSET], AX
	MOV	AX, SS:[BP+14]
	MOV	CS:[SI+PIXEL_DATA_SEGMENT], AX

	MOV	SP, BP
	POP	BP
	RET	12

SAVE_PARAMETERS ENDP

;***************************************************************************
;
	PUT_PIXEL PROC NEAR
;
;	set or reset a pixel at given coordinates in the printer buffer
;
;***************************************************************************

	PUSH	AX
	PUSH	BX
	PUSH	CX
	PUSH	DX
	PUSH	DI
	PUSH	SI

	CALL	GET_DATA_DISPLACEMENT

	OR	CX, 0000H
	JS	DONE2
	OR	DX, 0000H
	JS	DONE2
	CMP	DX, CS:[SI+NUM_LINES]
	JAE	DONE2
	CMP	CX, CS:[SI+NUM_PIXELS]
	JAE	DONE2

	CALL	GET_PIXEL_DATA_ADDR

	MOV	AX,CX
	SHR	AX,1
	SHR	AX,1
	SHR	AX,1
	MOV	BX,AX

	CALL	GET_BPL
	MUL	DX
	ADD	BX,AX
	MOV	AL,80H
	AND	CX,7
	SHR	AL,CL

	POP	DX
	PUSH	DX
	OR	DX, DX
	JZ	RESET
	OR	ES:[BX+DI], AL
	JMP	SHORT DONE2
RESET:
	NOT	AX
	AND	ES:[BX+DI], AL
DONE2:
	POP	SI
	POP	DI
	POP	DX
	POP	CX
	POP	BX
	POP	AX

	RET

PUT_PIXEL ENDP

;***************************************************************************
;
	GET_DATA_DISPLACEMENT PROC NEAR
;
;	calculate required displacement to address data
;
;***************************************************************************

	CALL	GET_IP

GET_IP:
	POP	SI
	SUB	SI, (GET_IP - BEGIN)

	RET

GET_DATA_DISPLACEMENT ENDP

;***************************************************************************
;
	GET_DRIVER_SETTINGS PROC NEAR
;
;***************************************************************************

	PUSH	BP
	PUSH	SI
	PUSH	DS
	PUSH	ES

	CALL	GET_DATA_DISPLACEMENT
	MOV	AL, CS:[SI+PARSED]
	OR	AL, AL
	MOV	CS:[SI+PARSED],1
	JNE	DONE5

	MOV	AH, 62H
	INT	21H

	PUSH	BX
	POP	DS
	MOV	AX, DS:[002CH]
	PUSH	AX
	POP	DS
	PUSH	CS
	POP	ES

	LEA	DI, [SI+ENVVAR]
	MOV	BX, SI
	XOR	SI, SI

TRAVERSE:
	CMP	BYTE PTR DS:[SI], 00H
	JZ	DONE5

	PUSH	DI
	MOV	CX, 7
	REPE	CMPSB
	POP	DI
	JNE	NEXT_VAR

NEXT_SWITCH:
	LODSB
	CMP	AL, 20H
	JE	NEXT_SWITCH
	OR	AL, AL
	JZ	DONE5

	DEC	SI
	MOV	AX, SI

CHECK_POS:
	LEA	DI, [BX+SWITCH_POS]
	MOV	SI, AX
	MOV	CX, 3
	REPE	CMPSB
	JNE	CHECK_NEG

	MOV	ES:[BX+POS], 1
	JMP	SHORT NEXT_SWITCH

CHECK_NEG:
	LEA	DI, [BX+SWITCH_NEG]
	MOV	SI, AX
	MOV	CX, 3
	REPE	CMPSB
	JNE	NEXT_SWITCH

	MOV	ES:[BX+POS], 0
	JMP	SHORT NEXT_SWITCH

NEXT_VAR:
	LODSB
	OR	AL, AL
	JNZ	NEXT_VAR
	JMP	SHORT TRAVERSE

DONE5:
	POP	ES
	POP	DS
	POP	SI
	POP	BP
	RET


GET_DRIVER_SETTINGS ENDP

;***************************************************************************
;
	GET_PIXEL_DATA_ADDR PROC NEAR
;
;	return segment and offset of the printer buffer
;
;***************************************************************************

	PUSH	SI

	CALL	GET_DATA_DISPLACEMENT
	MOV	ES, CS:[SI+PIXEL_DATA_SEGMENT]
	MOV	DI, CS:[SI+PIXEL_DATA_OFFSET]

	POP	SI
	RET

GET_PIXEL_DATA_ADDR ENDP

;***************************************************************************
;
	WRITE_HEADER PROC NEAR
;
;	generate corresponding PCX header and put onto output stream
;
;***************************************************************************

	CALL	GET_DATA_DISPLACEMENT

	MOV	AX, CS:[SI+HRES]
	DEC	AX
	MOV	CS:[SI+HDR_XMAX], AX

	MOV	AX, CS:[SI+VRES]
	DEC	AX
	MOV	CS:[SI+HDR_YMAX], AX

	CALL	GET_HDPI
	MOV	CS:[SI+HDR_HDPI], AX

	CALL	GET_VDPI
	MOV	CS:[SI+HDR_VDPI], AX

	CALL	GET_BPL
	INC	AX				; ROUND UP TO EVEN NUMBER
	AND	AX, 0FFFEH
	MOV	CS:[SI+HDR_BYTES_PER_LINE], AX

	XOR	AX, AX
	LEA	BX, [SI+HEADER]
	MOV	CX, PCX_HEADER_SIZE

WRITE:
	MOV	AL, CS:[BX]
	PUSH	AX
	CALL	WRITE_DATA
	INC	BX
	LOOP	WRITE

	RET

WRITE_HEADER ENDP

;***************************************************************************
;
	WRITE_RLE PROC NEAR
;
;***************************************************************************

	PUSH	BP
	MOV	BP, SP

	PUSH	AX
	PUSH	BX

	MOV	AX, SS:[BP+4]
	XOR	BX, BX

	OR	AH, AH
	JZ	DONE3
	CMP	AH, 01H
	JNE	SAVE_RUN
	CMP	AL, 0C0H
	JB	SAVE_SINGLE

SAVE_RUN:
	MOV	BL, AH
	ADD	BL, 0C0H
	PUSH	BX
	CALL	WRITE_DATA

SAVE_SINGLE:
	MOV	BL, AL
	PUSH	BX
	CALL	WRITE_DATA

DONE3:
	POP	BX
	POP	AX

	MOV	SP, BP
	POP	BP
	RET	2

WRITE_RLE ENDP

;***************************************************************************
;
	WRITE_DATA PROC NEAR
;
;	put a byte onto output stream
;
;***************************************************************************

	PUSH	BP
	MOV	BP, SP

	PUSH	AX
	PUSH	BX
	PUSH	CX
	PUSH	DX
	PUSH	ES
	PUSH	DI
	PUSH	SI

	CALL	GET_DATA_DISPLACEMENT

	PUSH	BP
	PUSH	SI
	PUSH	DS
	PUSH	ES

	MOV	AX, SS:[BP+4]
	PUSH	AX
	MOV	AX, CS:[SI+WRITE_DATA_ADDRESS]
	CALL	AX

	POP	ES
	POP	DS
	POP	SI
	POP	BP

	OR	AX, AX
	JZ	DONE4

	MOV	BP, CS:[SI+STACK_POINTER]
	MOV	SP, BP
	POP	ES
	POP	BP
	RET	4

DONE4:
	POP	SI
	POP	DI
	POP	ES
	POP	DX
	POP	CX
	POP	BX
	POP	AX

	MOV	SP, BP
	POP	BP
	RET	2

WRITE_DATA ENDP

;***************************************************************************

PATCH_SIZE	DB	(4096 - ($ - BEGIN)) DUP (00H)

CODE	ENDS
	END	BEGIN
